"""
Config is strictly identical between Keras and EpyNN.
Parameters in ### SHARED CONSTANTS ### are tunable and propagate to Keras and EpyNN.
Weights/bias are initialized with Keras and used for Keras and EpyNN.
See in ./utils.py for more code.

Architecture is:
    GRU > GRU > Dense
"""
####################### SHARED CONSTANTS ######################
# Data
N_SAMPLES = 32
STEPS = 10
VSIZE = 5

BATCH_SIZE = N_SAMPLES # We don't make any batch here

# Layers
gru0_units = 4
gru0_activation = 'tanh'
gru0_sequences = True
gru0_trainable = True
gru0_in = True

gru1_units = 2
gru1_activation = 'tanh'
gru1_sequences = False
gru1_trainable = True
gru1_in = True

dense_units = 2
dense_activation = 'softmax'
dense_trainable = False

# Run settings
N_EPOCHS = 2
LRATE = 0.01

LOSS = 'CCE'

# Support for autotest_GRU_Dense.py
import sys

if len(sys.argv) > 1:
    del sys.argv[0]
    args = sys.argv
    args = [int(arg) if arg.isdigit() else arg for arg in args]
    args = [True if arg == 'True' else arg for arg in args]
    args = [False if arg == 'False' else arg for arg in args]
    gru0_units, gru0_activation, gru0_sequences, gru0_trainable, gru0_in, gru1_units, gru1_activation, gru1_sequences, gru1_trainable, gru1_in, dense_units, dense_activation, dense_trainable, LOSS = args

######################## SEEDING/CFG ##########################
from timeit import default_timer as timer
from termcolor import cprint
import numpy as np
import random
import os

# CPU run - Disable GPU
os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
os.environ["CUDA_VISIBLE_DEVICES"] = ""
# Hide tensorflow debug info on call
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

# Seeding everywhere
import tensorflow as tf

SEED = 1
random.seed(SEED)
np.random.seed(SEED)
tf.random.set_seed(SEED)

# Float precision for TF (NumPy defaults to float64)
tf.keras.backend.set_floatx('float64')

# NumPy behavior on float arithmetics errors
np.seterr(all='raise')

######################## SHARED DATA ##########################
from utils import log_data

# Prepare dummy data
X_train = np.random.randint(-2, 2, (N_SAMPLES, STEPS, VSIZE)).astype('float64')
Y_train = np.random.randint(0, dense_units, (N_SAMPLES))

# One-hot encode X_train and Y_train
Y_train = np.eye(dense_units)[Y_train]

# Logs
log_data(X_train, Y_train)

###############################################################
##################### TENSORFLOW/KERAS ########################
from tensorflow.keras import Sequential
from tensorflow.keras.layers import Dense as KDense
from tensorflow.python.keras.layers.recurrent import GRU as KGRU
from tensorflow.keras.layers import Flatten as KFlatten
from utils import get_losses


keras_loss = get_losses(LOSS)
keras_opt = tf.keras.optimizers.SGD(learning_rate=LRATE)

# Layers
keras_gru0 = KGRU(gru0_units, activation=gru0_activation, return_sequences=gru0_sequences, recurrent_activation='sigmoid', implementation=1)
keras_gru1 = KGRU(gru1_units, activation=gru1_activation, return_sequences=gru1_sequences, recurrent_activation='sigmoid', implementation=1)

keras_flatten = KFlatten()
keras_dense = KDense(dense_units, activation=dense_activation)

# Model
keras_model = Sequential()
if gru0_in:
    keras_model.add(keras_gru0)
    keras_gru0.trainable = False
if gru1_in:
    keras_model.add(keras_gru1)
    keras_gru1.trainable = False

keras_model.add(keras_flatten)

keras_dense.trainable = dense_trainable
keras_model.add(keras_dense)

# Init
keras_model.compile(optimizer=keras_opt, loss=keras_loss, metrics=['accuracy'])
keras_model.fit(X_train, Y_train, epochs=N_EPOCHS, batch_size=BATCH_SIZE, verbose=1)

################# SHARED PARAMETERS

if gru0_in:
    gru0_Uz, gru0_Ur, gru0_Uhh = np.split(keras_gru0.get_weights()[0], 3, axis=1)
    gru0_Vz, gru0_Vr, gru0_Vhh = np.split(keras_gru0.get_weights()[1], 3, axis=1)
    gru0_bz, gru0_br, gru0_bhh = np.split(keras_gru0.get_weights()[2], 3, axis=0)

if gru1_in:
    gru1_Uz, gru1_Ur, gru1_Uhh = np.split(keras_gru1.get_weights()[0], 3, axis=1)
    gru1_Vz, gru1_Vr, gru1_Vhh = np.split(keras_gru1.get_weights()[1], 3, axis=1)
    gru1_bz, gru1_br, gru1_bhh = np.split(keras_gru1.get_weights()[2], 3, axis=0)

dense_W, dense_b = keras_dense.get_weights()
###################### RESUME
###################### TENSORFLOW/KERAS TRAINING
keras_gru0.trainable = gru0_trainable
keras_gru1.trainable = gru1_trainable
keras_dense.trainable = dense_trainable

keras_start = timer()
keras_model.compile(optimizer=keras_opt, loss=keras_loss, metrics=['accuracy'])
keras_model.fit(X_train, Y_train, epochs=N_EPOCHS, batch_size=BATCH_SIZE, verbose=1)
keras_end = timer() - keras_start
#############################
keras_model.summary()

keras_preds = keras_model.predict(X_train)

###############################################################
############################# EpyNN ###########################
from epynn.commons.maths import tanh
from epynn.dense.models import Dense as EDense
from epynn.gru.models import GRU as EGRU
from epynn.embedding.models import Embedding
from epynn.network.models import EpyNN
from epynn.flatten.models import Flatten as EFlatten
from utils import get_activation
from settings import se_hPars

se_hPars['learning_rate'] = LRATE
gru0_activation = get_activation(gru0_activation)
gru1_activation = get_activation(gru1_activation)
dense_activation = get_activation(dense_activation)

# Layers
epynn_embedding = Embedding(X_train, Y_train, relative_size=(1, 0, 0), batch_size=BATCH_SIZE)
epynn_gru0 = EGRU(unit_cells=gru0_units, sequences=gru0_sequences, activate=gru0_activation, clip_gradients=False)
epynn_gru1 = EGRU(unit_cells=gru1_units, sequences=gru1_sequences, activate=gru1_activation, clip_gradients=False)

epynn_flatten = EFlatten()
epynn_dense = EDense(units=dense_units, activate=dense_activation)

# Model
layers = [epynn_embedding]

if gru0_in:
    epynn_gru0.p['Uz'] = gru0_Uz
    epynn_gru0.p['Vz'] = gru0_Vz
    epynn_gru0.p['bz'] = gru0_bz
    epynn_gru0.p['Ur'] = gru0_Ur
    epynn_gru0.p['Vr'] = gru0_Vr
    epynn_gru0.p['br'] = gru0_br
    epynn_gru0.p['Uhh'] = gru0_Uhh
    epynn_gru0.p['Vhh'] = gru0_Vhh
    epynn_gru0.p['bhh'] = gru0_bhh
    epynn_gru0.trainable = gru0_trainable
    layers.append(epynn_gru0)

if gru1_in:
    epynn_gru1.p['Uz'] = gru1_Uz
    epynn_gru1.p['Vz'] = gru1_Vz
    epynn_gru1.p['bz'] = gru1_bz
    epynn_gru1.p['Ur'] = gru1_Ur
    epynn_gru1.p['Vr'] = gru1_Vr
    epynn_gru1.p['br'] = gru1_br
    epynn_gru1.p['Uhh'] = gru1_Uhh
    epynn_gru1.p['Vhh'] = gru1_Vhh
    epynn_gru1.p['bhh'] = gru1_bhh
    epynn_gru1.trainable = gru1_trainable
    layers.append(epynn_gru1)

layers.append(epynn_flatten)

epynn_dense.p['W'] = dense_W
epynn_dense.p['b'] = dense_b
epynn_dense.trainable = dense_trainable
layers.append(epynn_dense)

# Init
epynn_model = EpyNN(layers=layers)
epynn_model.initialize(loss=LOSS, se_hPars=se_hPars.copy(), seed=SEED, params=False, end='\r')

### EpyNN TRAINING ###
epynn_start = timer()
epynn_model.train(epochs=N_EPOCHS, verbose=1, init_logs=False)
epynn_end = timer() - epynn_start
######################

epynn_preds = epynn_model.forward(X_train)


###############################################################
############################ Compare ##########################
from utils import print_logs

data = (Y_train, epynn_preds, keras_preds, LOSS, keras_loss, epynn_model, keras_end, epynn_end)

kloss_RMSD, preds_RMSD = print_logs(data)

if len(sys.argv) > 1:
    args = [str(arg) for arg in args]
    msg = str(N_SAMPLES)+', '+str(STEPS)+', '+str(VSIZE)+', '+', '.join(args)+', '+str(kloss_RMSD)+', '+str(preds_RMSD)
    with open('autotest_GRU_Dense.dat', 'a') as f:
        f.write(msg)
        f.write('\n')
